#!/usr/bin/env python

import time, sys, os, signal, re, cPickle
import OSC

class Grid:

    _state = {}
    
    _client = None # OSC client
    _pid = None # PID of running interpreter

    _cycles = 0


    def __init__(self, cols, rows):
        """ Initialize a grid, with rows and cols """
        
        self._state['rows'] = rows;
        self._state['cols'] = cols;
        self._state['data'] = [0] * self._state['rows']


    def from_file(fn):
        """ Initialize a grid from an image file. Static method: the
        image dimensions form the grid dimensions."""

        from PIL import Image
        im = Image.open(fn)
        w, h = im.size

        g = Grid(w, h)
        g.clear()
        g.fill_from_image(im)
        return g
    from_file = staticmethod(from_file)


    def to_file(self, filename, cellwidth=10, fg=(255,255,255), bg=(0,0,0)):

        from PIL import Image, ImageDraw
        im = Image.new("RGB", (self._state['cols']*cellwidth, self._state['rows']*cellwidth), bg)

        draw = ImageDraw.Draw(im)
        
        for x in range(self._state['cols']):
            for y in range(self._state['rows']):
                if self.get(x, y):
                    fill = fg
                else:
                    fill = bg
                draw.rectangle( ( (x*cellwidth, y*cellwidth), ((x+1)*cellwidth, (y+1)*cellwidth)), fill=fill)
        im.save(filename, filename.split(".")[-1].upper())
        


    def fill_from_image(self, im):
        """ Given a PIL image, fill the current grid with the
        image. Black pixels are left empty: non-black pixels
        give a"filled" pixel. Note: no bounds checking is
        done on the image! """

        for x in range(self._state['cols']):
            for y in range(self._state['rows']):
                val = im.getpixel((x,y))
                if type(val) == tuple:
                    val = sum(val)
                                  
                if val > 0:
                    self.set(x, y, True)
        pass

    def resize(self, w, h):
        self._state['cols'] = w
        self._state['rows'] = h

    def _warp(self, x, y):

        """ Given a coordinate, 'warp' it so it falls into the
        grid. Can be overridden to get different grid behaviour."""

        return x % self._state['cols'], y % self._state['rows']

    def get(self, x, y):
        """ Get the pixel at (x,y). Returns TRUE when filled, or FALSE when empty. """
        x, y = self._warp(x, y)
        return self._state['data'][y] & (1<<x)


    def set(self, x, y, val):
        """ Set the pixel at (x,y) to val (TRUE or FALSE). """
        x, y = self._warp(x, y)

        if val == True:
            self._state['data'][y] = self._state['data'][y] | (1<<x)
        else:
            self._state['data'][y] = self._state['data'][y] & (((1<<self._state['cols'])-1) ^ (1<<x))

        self._cycles += 1
        if self._cycles > 1000:
            self.sync()


    def clear(self):
        """ Clear the grid -- all pixels black. """
        self._state['data'] = [0] * self._state['rows']


    def invert(self):
        """ Invert all pixels in the grid. """
        self._state['data'] = [((1<<self._state['cols'])-1) ^ c for c in self._state['data']]


    def fill(self):
        """ Fill the grid. """
        self.clear()
        self.invert()


    def __str__(self):
        """ Represent this grid as a string, for debugging and such. """
        s = "Grid %dx%d:\n" % (self._state['cols'], self._state['rows'])
        for y in range(self._state['rows']):
            for x in range(self._state['cols']):
                if self.get(x, y):
                    s += "X"
                else:
                    s += "."
            s += "\n"
        return s


    def get_size(self):
        """ Return the size tuple of this grid (cols, rows). """
        return self._state['cols'], self._state['rows']


    def get_cols(self):
        """ Return the number of columns in this grid. """
        return self._state['cols']


    def get_rows(self):
        """ Return the number of rows in this grid. """
        return self._state['rows']


    def run(self, str):
        """ Run a grid script """
        tokens = [x.strip() for x in re.split(r'[\s]+', str) if len(x) > 0]
        tokens.reverse()

        args = []
        cmd = []
        v = 0
        for t in tokens:
            if len(t) == 0:
                continue
            
            try:
                v = int(t)
                # int argument, push to args
                args.append(v)
                continue

            except ValueError:
                pass

            x = getattr(self, t)
            cmd.append( (t, args) )
            args = []

        cmd.reverse()

        pid = os.fork()
        if pid == 0:
            signal.signal(signal.SIGTERM, self.sigterm)
            while True:
                for c,arg in cmd:
                    getattr(self,c)(*arg)
            pass

        self._pid = pid

        tokens.reverse()
        self._send_tokens(tokens)


    def sigterm(self, sig, frame):
        self.save_state()
        sys.exit(0)

        
    def save_state(self, sig=None, frame=None):
        fn = self._get_state_file()

        while os.path.exists(fn + ".lock"):
            time.sleep(0.01)
        l = open(fn + ".lock", "w")
        
        try:
            f = open(self._get_state_file(), 'w')
        except IOError:
            print "Unable to save grid state"
            return

        cPickle.dump(self._state, f)
        f.close()
        l.close()
        os.unlink(fn + ".lock")
        



    def load_state(self):
        fn = self._get_state_file()

        while os.path.exists(fn + ".lock"):
            time.sleep(0.01)
        
        try:
            f = open(fn, 'r')
        except IOError:
            print "Unable to load grid state"
            return

        self._state = cPickle.load(f)


    def _get_state_file(self):
        return os.getcwd() + os.sep + ".grid_state"


    def stop(self):
        """ Stop a running grid interpreter """
        
        assert self._pid is not None, "Grid is not running"
        os.kill(self._pid, signal.SIGTERM)
        self._pid = None

    def running(self):
        return self._pid is not None


    def _send_tokens(self, tokens):
        """ Send the code to the display for vis """
        self._init_osc()
        msg = OSC.OSCMessage("/tokens")

        msg.append(len(tokens))
        for t in tokens:
            msg.append(str(t))
        self._client.send(msg)


    def sync(self):
        """ Sync grid with display / visualization """
        
        self._init_osc()
        msg = OSC.OSCMessage("/grid")
        msg.append(self._state['cols'])
        msg.append(self._state['rows'])
        for i in self._state['data']:
            msg.append(i)
        self._client.send(msg)

        self._cycles = 0


    def wait(self, x=0, *ign):
        self.sync()
        time.sleep(0.025 * (1<<x))


    def _init_osc(self):
        if self._client is None:
            self._client = OSC.OSCClient()
            self._client.connect(("127.0.0.1", 2200))




if __name__ == "__main__":

    g = Grid.from_file("test.png")
    print g
    sys.exit(1)
    w = 16
    g = Grid1(w,w)
    g.run("down east 4 wait 2 south 3 north wait 2")
    time.sleep(2)
    g.stop()
    
    print("bye")
    
