// -------------------------------------------------------------
// eeprom.cpp
//

#include <Wire.h>
#include <WProgram.h>

#include "eeprom.h"


// Low-level EEPROM read functions.
void i2c_eeprom_write_byte( int deviceaddress, unsigned int eeaddress, byte data )
{
    int rdata = data;
    Wire.beginTransmission(deviceaddress);
    Wire.send((int)(eeaddress >> 8)); // MSB
    Wire.send((int)(eeaddress & 0xFF)); // LSB
    Wire.send(rdata);
    Wire.endTransmission();
	delay(4); // give the chip some time to recover!
}

// WARNING: address is a page address, 6-bit end will wrap around
// also, data can be maximum of about 30 bytes, because the Wire library has a buffer of 32 bytes
void i2c_eeprom_write_page( int deviceaddress, unsigned int eeaddresspage, byte* data, byte length ) 
{
    Wire.beginTransmission(deviceaddress);
    Wire.send((int)(eeaddresspage >> 8)); // MSB
    Wire.send((int)(eeaddresspage & 0xFF)); // LSB
    byte c;
    for ( c = 0; c < length; c++) {
	Wire.send(data[c]);
    }
    Wire.endTransmission();
    delay(4); // give the chip some time to recover!
}

byte i2c_eeprom_read_byte( int deviceaddress, unsigned int eeaddress ) 
{
    byte rdata = 0xFF;
    Wire.beginTransmission(deviceaddress);
    Wire.send((int)(eeaddress >> 8)); // MSB
    Wire.send((int)(eeaddress & 0xFF)); // LSB
    Wire.endTransmission();
    Wire.requestFrom(deviceaddress,1);
    if (Wire.available()) rdata = Wire.receive();
    return rdata;
}

// let's not read more than 30 or 32 bytes at a time; 
// there seems to be an issue in the wire library...
void i2c_eeprom_read_buffer( int deviceaddress, unsigned int eeaddress, byte *buffer, int length ) 
{
    Wire.beginTransmission(deviceaddress);
    Wire.send((int)(eeaddress >> 8)); // MSB
    Wire.send((int)(eeaddress & 0xFF)); // LSB
    Wire.endTransmission();
    Wire.requestFrom(deviceaddress,length);
    int c = 0;
    for ( c = 0; c < length; c++ ) {
	if (Wire.available()) buffer[c] = Wire.receive();
    }
}

// end lowlevel functions

// initialize an empty eeprom structure
EEPROM *eeprom_init(int addr)
{
    EEPROM *e = (EEPROM *)malloc(sizeof(EEPROM));

    e->addr = addr;
    e->cur  = 0;    // set current position to start of the chip
    e->max = MAX_ADDR; // The maximum address
    return e;
}


// read a line from the EEPROM module, until we read either 120 chars,
// or read a newline (character 10), or the end-of-memory marker(character 0).
// memory should be free()'d by the caller!!!! 
char *eeprom_read_line(EEPROM *e)

{
    byte buf[PAGE_SIZE];

    // output buffer..
    char *b = (char *)malloc(120 * sizeof(char));

    int i; // byte counter
    int s; // how much bytes are we gonna read?
    int cont = 1; // continue flag
    int read = 0; // counter, how much have we read

    while (cont)
    {
	//s = e->cur + PAGE_SIZE < e->max ? PAGE_SIZE : (e->max - e->cur);
	//if (s < 0) break;
	s = PAGE_SIZE;
	
	i2c_eeprom_read_buffer(e->addr, e->cur, buf, s);

	for (i=0; i<s; i++) {
	    // loop over all read bytes; we need to check if we encounter any markers.
	    if (buf[i] == 10) {
		// encountered end-of-line, lets stop reading
		b[read++] = 0; // mark end-of-string
		cont = 0;
		i++; // progress one so we'll skip over the newline the next time
		break;
	    }
	    else if (buf[i] == 0) {
		// encountered end-of-memory
		b[read++] = 0; // mark end of string
		cont = 0;
		i = -1; // restart address
		break;
	    }
	    else {
		// add the read char to the buffer
		b[read++] = buf[i];
	    }
	}
	if (read >= 120) cont = 0;

	if (i >= 0) {
	    // advance current-location pointer
	    e->cur += i;
	} else {
	    // restart
	    e->cur = 0;
	}
    }
    
    return b;
}
