import pygame

from twisted.application import service
from twisted.internet    import task

import btg.text


class TextGUI(service.Service):
    """ A GUI component in the form of a window with Arial letters representing the monitors. """
    
    screen_w = 20
    screen_h = 15

    def startService(self):
        service.Service.startService(self)

        pygame.display.init()
        pygame.font.init()
        pygame.display.set_mode( (self.screen_w * self.parent.screens, 3*self.screen_h ))

        df = "/usr/share/fonts/truetype/msttcorefonts/arial.ttf"
        try:
            self.font = pygame.font.Font(df, self.screen_h);
        except Exception: # Font not found?
            pass # do nothing

        lc = task.LoopingCall(self.tick)
        lc.start(0.05)

    def on_parent_resize(self, newsize):
        pygame.display.set_mode( (self.screen_w * newsize, 3*self.screen_h ))
        

    def stopService(self):

        pygame.font.quit()
        pygame.display.quit()


    def tick(self):
        screen = pygame.display.get_surface()

        screen.fill( (0,0,0) )

        x = 0
        for letter in self.parent.state:
            s = self.font.render(letter, True, (255,255,255))
            screen.blit(s, (x*self.screen_w, self.screen_h))
            x += 1
            
        pygame.display.flip()



class BitmappedGUI(service.Service):
    """ A GUI component in the form of a window with bitmapped letters representing the monitors. """

    bitmaps = {}  # Bitmap-cache
    font = None   # The font instance to use
    
    screen_w = None # width and height are loaded from the bitmaps
    screen_h = None

    padding = 10
    bgcolor = (128,128,128)

    def __init__ (self, font):
        self.font = btg.text.Font(font)

        import os
        self.dir  = os.path.dirname(font) + "/out/" + os.path.basename(font).split(".")[0] + "/"

    def on_parent_resize(self, newsize):
        pygame.display.set_mode( ((self.screen_w + 2*self.padding) * newsize + 2*self.padding, 2*self.padding + self.screen_h ))
        pygame.display.set_caption("vga_line")

    def loadBitmaps (self):

        import glob, os

        for f in glob.glob(self.dir + "*.png"):
            bitmap = pygame.image.load(f)
            letter = chr(int(os.path.basename(f).split(".")[0]))
            self.bitmaps[letter] = bitmap

            self.screen_w = bitmap.get_width()
            self.screen_h = bitmap.get_height()

        self.padding  = self.screen_w / 10
        
        missing = set(self.font.letters()).difference(set(self.bitmaps.keys()))
        if missing:
            print "Missing characters in bitmaps!!!", missing


    def startService(self):
        service.Service.startService(self)

        self.loadBitmaps()

        pygame.display.init()
        pygame.font.init()
        pygame.display.set_mode( ((self.screen_w + 2*self.padding) * self.parent.screens + 2*self.padding, 2*self.padding + self.screen_h ))
        pygame.display.set_caption("vga_line")

        lc = task.LoopingCall(self.tick)
        lc.start(0.05)

    def stopService(self):

        pygame.font.quit()
        pygame.display.quit()


    def tick(self):
        screen = pygame.display.get_surface()

        screen.fill( self.bgcolor )

        x = 0
        for letter in self.parent.state:

            if letter not in self.font.letters():
                print "Missing character in fontset!!!", ord(letter)
                letter= "?"
                #x += 1
                #continue

            bmp = self.bitmaps[letter]

            startx = self.padding + x*(self.screen_w+2*self.padding)
            screen.blit(bmp, (startx, self.padding))

            pygame.draw.line(screen, (0,0,0), (startx+self.screen_w+self.padding-1, 0), (startx+self.screen_w+self.padding-1, 2*self.padding+self.screen_h), 2)
            
            x += 1
            
        pygame.display.flip()

