#!/usr/bin/env python
#
# Simulator for Multi-VGA monitor lineup. Mostly used as a playground
# to experiment with different scrolling techniques, speeds,
# directions, animation styles, etc, etc.
#
# May 2009, Arjan Scherpenisse <arjan@scherpenisse.net>
#
# This file is released under the MIT license.
#

from twisted.application import service
from twisted.internet    import reactor, task, defer

import random, sys
sys.path.append('/home/arjan/devel/final/_oneliner')

import input, gui  # My custom components


class LineLogic(service.MultiService):
    """ Line logic -- this manages the data which is displayed on the line. """

    # The number of screens we have
    screens = None

    # The current state: a string of length <self.screens>.
    state   = None

    def __init__(self, screens=76):
        service.MultiService.__init__(self)

        # Initialize the logic
        self.resize(screens)

    def resize(self, size):
        """ Resizes the state, and resets it. """
        self.screens = size
        self.state   = "-" * self.screens
        for child in self:
            if getattr(child, 'on_parent_resize', False):
                child.on_parent_resize(size)
        
            

    def startService(self):
        service.MultiService.startService(self)


    def set(self, x, str, wrap=True):
        """ Overwrite the state at pos x. All characters will be written at once. """
        # TODO -- make the special \0 character be a "transparent" space.
        # TODO -- other control chars?

        # normalize the X argument
        while x < 0: x += self.screens
        while x >= self.screens: x -= self.screens

        l = len(str)

        if x + l <= self.screens:
            # No wrapping, the easy case.
            self.state = self.state[0:x] + str + self.state[x+l:]
        else:
            # Wrapping/clipping.
            # Chop str off at self.screens-x
            o = self.screens-x
            
            if not wrap:
                self.state = self.state[0:x] + str[0:o]
            else:
                self.state = str[o:] + self.state[l-o:x] + str[0:o]

        self.state = self.state.upper()
        pass

    def clear(self):
        self.state = " " * self.screens


    def add_scroller(self, text, offset = 0, ttl = None, speed = 0.1):
        while offset < 0: offset += self.screens
        while offset >= self.screens: offset -= self.screens

        scr = Scroller()
        scr.offset = offset
        scr.text = text
        if speed < 0:
            scr.forward = False
            speed = -speed
        scr.speed = speed
        scr.setServiceParent(self)

        if ttl:
            return scr.set_ttl(ttl)
        return None
        



class BaseAnimation (service.Service):

    speed = 0.2
    _lc = None
    
    def startService(self):
        service.Service.startService(self)

        self._lc = task.LoopingCall(self.tick)
        self._lc.start(self.speed)

    def tick(self):
        pass

class LineCycler (BaseAnimation):

    speed = .03
    i = 0
    
    def __init__(self, lines):
        self.lines = lines
        
    def tick(self):
        self.parent.clear()
        s = self.lines[self.i]
        #s = s.strip()
        s = s.replace("\t", "    ")
        s = s[:self.parent.screens]
        if s[-1:] == "\n":
            s = s[:-1] # strip the newline
        self.parent.set(0, s)
        self.i = (self.i + 1) % len(self.lines)
        

class Scroller (BaseAnimation):

    speed   = 0.1               # How fast
    forward = True              # Forward or backward?
    offset  = 0                 # Where to start the text        
    text    = "A simple text."  # Which text to scroll
    ttl     = None              # how many ticks are we supposed to survive?
    ttl_deferred = None         # Deferred to fire after ttl has gone
    
    _i = 0

    def tick(self):

        self.parent.set(self.offset + self._i, " " + self.text + " ")
        if self.forward:
            self._i -= 1
        else:
            self._i += 1
    
        if type(self.ttl) == int:
            self.ttl -= 1
            if self.ttl <= 0:
                self._lc.stop()
                self.disownServiceParent()
                self.ttl_deferred.callback(self.offset + self._i)


    def set_ttl(self, ttl):
        """ Set the Ticks-to-live for this scroller """
        self.ttl = ttl
        if ttl:
            self.ttl_deferred = defer.Deferred()
        else:
            self.ttl_deferred = None
            
        return self.ttl_deferred


class TBT (service.Service):

    pairs = []
    x = 0
    
    def __init__(self, file):
        from struct import unpack
        
        fp = open(file, "r")
        while True:
            data = fp.read(16)
            if not data:
                break

            (char,l1,l2) = unpack("c7x2L", data)
            self.pairs.append( (char, l1) )

        print self.pairs

    def startService(self):

        t = 0
        for char, time in self.pairs:
            t += time
            reactor.callLater(t/1000.0, self.letter, char)
                
    def letter(self, l):
        if (ord(l) == 13):
            self.x = 0
            self.parent.clear()
            return

        if (ord(l) == 9):
            l = "  "
            
        self.x += len(l)
        self.parent.set(self.x, l)
        

application = service.Application("one_liner simulator")

logic = LineLogic(64)
logic.setServiceParent(application)

g = gui.BitmappedGUI("../fonts/5x7.txt")
g.setServiceParent(logic)


import glob

try:
    serialport = glob.glob("/dev/ttyUSB*")[0]
except Exception:
    serialport = None

if serialport:
    # run the simulator with the Arduino as the controller of the display
    p = input.ControllerInput(serialport, baudrate=115200, timeout=1)
    p.setServiceParent(logic)

else:
    # Run the simulator with this program as the controller; using some fancy anymation-prototyping stuff

    # A simple source-code scroller
    def source_code(f, o=0):
        lines = [s.strip().replace("\t", " ") for s in open(f, "r").readlines() if len(s.strip())]
        def mk(i):
            if i % 2:
                speed = 0.15
            else:
                speed = -0.15
            ttl = random.randint(20, 100)

            logic.add_scroller(random.choice(lines), i, ttl, speed).addCallback(mk)

        mk(0)

    #tbt = TBT("hello.tbt")
    #tbt.setServiceParent(logic)


    #source_code("../generate/vga.asm", 40)
    #source_code(__file__)

    #logic.add_scroller("..", 0, None, 0.3)
    #logic.add_scroller("..$", 20, None, -0.3)

    #c = LineCycler(open(__file__, "r").readlines())
    #c.setServiceParent(logic)

    #c = LineCycler(open("/home/arjan/devel/final/final_work/controller/animate.cpp", "r").readlines()
    c = LineCycler(open("../comments", "r").readlines())
    c.speed = 1
    c.setServiceParent(logic)
    
    #c = LineCycler(open("/home/arjan/devel/final/final_work/vga-letter/vga-letter.asm", "r").readlines())
    #c.setServiceParent(logic)
    
